#define coder   "Beppu"

;
;Directive
;Label         Op     Operand        Comments
;                                 
;

#define b       byte
#define w       word
#define d       dword

; d a t a    s t r u c t u r e s ___________________________________________

; [Bresenham]
;
; infoB {
;   long        index   // direction of change  0BGR  aka jumpTable Index [0,4]
;                       // [3=0][2=b][1=g][0=r]
;   long        dR      // midpoint             R
;   long        incMAJR // major increment
;   long        incMINR // minor increment
;   long        dG      // midpoint             G
;   long        incMAJG // major increment
;   long        incMING // minor increment
;   long        dB      // midpoint             B
;   long        incMAJB // major increment
;   long        incMINB // minor increment
; }

        /* This should look familiar. */

; [DDA]
;
; infoD {
;   long        sign    // direction of change  0BGR  aka signTable Index [0,4]
;
;   long        fracR   // fractional part      R  // of course
;   long        ri      // change (Int)
;   long        rf      // change (Frac)
;   long        fracG   // fractional part      G  // the integer part
;   long        gi      // change (Int)
;   long        gf      // change (Frac)
;   long        fracB   // fractional part      B  // is the palette itself
;   long        bi      // change (Int)
;   long        bf      // change (Frac)
; }

; [GENERAL]
;
; palette {
;   char        r
;   char        g
;   char        b
; }

[bits 32]

; Linux

[global pal_update]
[global pal_initD]
[global pal_nextD]
[global pal_8to32]

; DJGPP

[global _pal_update]
[global _pal_initD]
[global _pal_nextD]
[global _pal_8to32]

[section .text]

/*

void    null (
        );

%% :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

null:
              ; retn


/*

Because Quill uses 32 bits for each RGB entry (for a total of 12 bytes
per color entry) this routine is needed to fatten palettes from 3 bytes
to 3 words.

void pal_8to32 (
    u32 *dst,
    u8 *src,
    u32 qty
    );

%% :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

pal_8to32:
_pal_8to32:

; frame takes into account local stack space up to and including the
; return address

%define frame   16
%define dst     esp+frame+0
%define src     esp+frame+4
%define qty     esp+frame+8

                ;: blah

                push    ebx
                push    edi
                push    esi

                mov     esi,[src]
                mov     edi,[dst]
                mov     ecx,[qty]

                xor     eax,eax
                sub     ebx,ebx
                xor     edx,edx
        .werk:
                mov     al,[esi]
                mov     bl,[esi+1]
                shl     eax,2
                mov     dl,[esi+2]
                shl     ebx,2
                add     esi,3
                shl     edx,2
                mov     [edi],eax
                mov     [edi+4],ebx
                mov     [edi+8],edx
                add     edi,12
                dec     ecx
                jnz     .werk

                pop     esi
                pop     edi
                pop     ebx
                
                retn


/* Update VGA hardware palette

void pal_update (
    char* pal,
    u32 x,
    u32 y
    );

%% :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

pal_update:
_pal_update:    push    esi

                mov     cl,00001001b
                mov     edx,3dah
        .wait:
                in      al,dx
                and     al,cl
                cmp     al,cl
                jne     .wait           ; Wait until vertical retrace.

                mov     ecx,[esp+16]
                mov     edx,0x3c8
                mov     eax,[esp+12]
                mov     esi,[esp+8]
                sub     ecx,eax
                inc     ecx

                out     dx,al

                inc     edx
                lea     ecx,[ecx+ecx*2]
                rep     outsb

                pop     esi
                retn


/* Initialize variables for DDA palette transitions.

void pal_initD (
    char* dst,
    char* src,
    u32 x,
    u32 y,
    u32 z,
    infoD* nfo
    );

%% :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

#define frame   16

#define dst     esp+frame+04
#define src     esp+frame+08
#define x       esp+frame+12
#define y       esp+frame+16
#define z       esp+frame+20
#define nfo     esp+frame+24

pal_initD:
_pal_initD:
                push    ebx
                push    ebp
                push    edi
                push    esi

                mov     ecx,[y]
                mov     eax,[x]
                mov     esi,[src]
                sub     ecx,eax
                mov     ebx,[dst]
                lea     eax,[eax+eax*2]
                mov     edi,[nfo]       ; edi = nfo
                add     esi,eax         ; esi = &src[x*3]
                add     ebx,eax         ; ebx = &dst[x*3]
                inc     ecx             ; ecx = y - x + 1 ` loop counter
;

; eax =
; ebx = &dst[x*3]
; ecx = y - x + 1
; edx =
; edi = nfo
; esi = &src[x*3]
; ebp =

.werk:          ; Red

                xor     edx,edx
                xor     eax,eax
                mov     dl,b [esi]      ; src
                mov     al,b [ebx]      ; dst
                mov     ebp,[edi]       ; naive P5 cache opt
                sub     eax,edx
                sbb     ebp,ebp         ; ebp = mask [0,-1]

                mov     edx,ebp
                idiv    d [z]
                mov     [edi+8],eax     ; ri
                xor     edx,ebp
                mov     eax,ebp
                ;
                rcl     ebp,1           ; ebp -1 : overflow
                ;
                adc     edx,b 0
                cmp     eax,b 1
                adc     eax,b 0
                mov     [edi+12],edx
                mov     [edi],al

                ; Green

                xor     edx,edx
                xor     eax,eax
                mov     dl,b [esi+1]    ; src
                mov     al,b [ebx+1]    ; dst
                mov     ebp,[edi+32]    ; next cache line
                sub     eax,edx
                sbb     ebp,ebp

                mov     edx,ebp
                idiv    d [z]
                mov     [edi+20],eax
                xor     edx,ebp
                mov     eax,ebp
                rcl     ebp,1
                adc     edx,b 0
                cmp     eax,b 1
                adc     eax,b 0
                mov     [edi+24],edx
                mov     [edi+1],al

                ; Blue

                xor     edx,edx
                xor     eax,eax
                mov     dl,b [esi+2]    ; src
                mov     al,b [ebx+2]    ; dst
                sub     eax,edx
                sbb     ebp,ebp

                mov     edx,ebp
                idiv    d [z]
                mov     [edi+32],eax
                xor     edx,ebp
                mov     eax,ebp
                rcl     ebp,1
                adc     edx,b 0
                cmp     eax,b 1
                adc     eax,b 0
                mov     [edi+36],edx
                add     esi,b 3
                add     ebx,b 3
                mov     [edi+2],al
                add     edi,b 40

                dec     ecx
                jnz     near .werk      ; 16 bytes too long <!#$!##@%>
;

                // init frac[R,G,B] to ZER0

                mov     ecx,[y]
                mov     eax,[x]
                mov     edi,[nfo]
                sub     ecx,eax
                inc     ecx
                xor     eax,eax
.werk1:
;
                mov     [edi+04],eax
                mov     [edi+16],eax
                mov     [edi+28],eax
                add     edi,40
                dec     ecx
                jnz     short .werk1
;
                pop     esi
                pop     edi
                pop     ebp
                pop     ebx

                retn

#undef  frame

#undef  dst
#undef  src
#undef  x
#undef  y
#undef  z
#undef  nfo



[section .text]

/* Calculate next palette (DDA)

void pal_nextD (
    char* dst,
    char* src,
    u32 x,
    u32 y,
    u32 z,
    infoD* nfo
    );

%% :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */

#define frame   16

#define dst     esp+frame+04
#define src     esp+frame+08
#define x       esp+frame+12
#define y       esp+frame+16
#define z       esp+frame+20
#define nfo     esp+frame+24

pal_nextD:
_pal_nextD:     push    ebx
                push    ebp
                push    edi
                push    esi

                mov     ecx,[y]
                mov     eax,[x]
                sub     ecx,eax
                mov     esi,[src]
                mov     edi,[nfo]
                mov     ebx,[dst]
                lea     eax,[eax+eax*2]
                inc     ecx
                add     esi,eax
                add     ebx,eax
.werk:
;
; eax =
; ebx = &dst[x*3]
; ecx = y - x + 1       ; [y] is used to store ecx temporarily
; edx =
; edi = nfo
; esi = &src[x*3]
; ebp =

                mov     [y],ecx

                ; red

                movsx   eax,b [edi]     ; ri.adjustment [-1,+1]
                mov     ebp,[edi+12]    ; rf
                mov     edx,[edi+4]     ; fracR
                add     edx,ebp
                mov     ebp,[z]
                dec     ebp
                cmp     ebp,edx
                mov     [edi+4],edx
                sbb     ecx,ecx
                inc     ebp
                and     eax,ecx         ; eax = ri.adjustment
                sub     edx,ebp
                mov     ebp,[edi+4]
                and     edx,ecx
                xor     ecx,b -1
                and     ebp,ecx
                or      edx,ebp
                mov     [edi+4],edx     ; [edi+4] = fracR (final)

                xor     ecx,ecx
                mov     cl,[esi]        ; R
                mov     edx,[edi+8]     ; ri
                add     edx,ecx
                add     eax,edx         ; R (final)
                mov     [ebx],al

                ; green

                movsx   eax,b [edi+1]   ; gi.adjustment [-1,+1]
                mov     ebp,[edi+24]    ; gf
                mov     edx,[edi+16]    ; fracG
                add     edx,ebp
                mov     ebp,[z]
                dec     ebp
                cmp     ebp,edx
                mov     [edi+16],edx    ; fracG (temp storage)
                sbb     ecx,ecx
                inc     ebp
                and     eax,ecx         ; eax = gi.adjustment
                sub     edx,ebp
                mov     ebp,[edi+16]
                and     edx,ecx
                xor     ecx,b -1
                and     ebp,ecx
                or      edx,ebp
                mov     [edi+16],edx    ; [edi+16] = fracG (final)

                xor     ecx,ecx
                mov     cl,[esi+1]      ; G
                mov     edx,[edi+20]    ; gi
                add     edx,ecx
                lea     ecx,[edx+eax]   ; G (final)
                mov     [ebx+1],cl

                ; blue

                movsx   eax,b [edi+2]   ; bi.adjustment [-1,+1]
                mov     ebp,[edi+36]    ; bf
                mov     edx,[edi+28]    ; fracB
                add     edx,ebp
                mov     ebp,[z]
                dec     ebp
                cmp     ebp,edx
                mov     [edi+28],edx    ; fracB (temp storage)
                sbb     ecx,ecx
                inc     ebp
                and     eax,ecx
                sub     edx,ebp
                mov     ebp,[edi+28]
                and     edx,ecx
                xor     ecx,b -1
                and     ebp,ecx
                or      edx,ebp
                mov     [edi+28],edx    ; [edi+28] = fracB (final)

                xor     ecx,ecx
                mov     cl,[esi+2]      ; B
                mov     edx,[edi+32]    ; bi
                add     edx,ecx
                lea     ecx,[edx+eax]   ; B (final)
                mov     [ebx+2],cl

                mov     ecx,[y]
                add     ebx,b 3
                add     esi,b 3
                add     edi,b 40
                dec     ecx
                jnz     near .werk
;

#undef  dst
#undef  src
#undef  x
#undef  y
#undef  z
#undef  nfo

.bye            pop     esi
                pop     edi
                pop     ebp
                pop     ebx

                retn
